const rewire = require('rewire')
const now = new Date('1980-03-18T04:15:00')

const util = require('util')
util.promisify = jest.fn().mockImplementation(func => func)

const fs = require('fs')
fs.readdir = jest.fn()
fs.stat = jest.fn()

const os = require('os')
os.hostname = jest.fn().mockImplementation(() => 'fakehost')

const mockSpawn = jest.fn()
const mockSend = jest.fn()
function PythonWorker (args) {
  this.spawn = mockSpawn
  this.send = mockSend
}
const interfaces = rewire('../../src/interfaces')
interfaces.__set__('PythonWorker', PythonWorker)

/**
 * Asynchronous mock function that returns a promise witch
 * resolves or rejects based on the arguments given.
 * @param {object} unnamed - Either error or result
 */
function AsyncMock ({ result, error }) {
  return new Promise((resolve, reject) => {
    if (error) reject(error)
    resolve(result)
  })
}

const { resolvers, typeDefs, __test__ } = interfaces

describe('interfaces module', () => {
  it('exports resolvers and typeDefs', () => {
    expect(resolvers).toBeDefined()
    expect(typeDefs).toBeDefined()
    expect(resolvers).toHaveProperty('Query'),
    expect(resolvers).toHaveProperty('Mutation')
  })

  it('finds workers', async () => {
    // Check that all possible return values of fs.readdir are correctly handled.
    const findWorkers = interfaces.__get__('findWorkers')

    // 1. Check use case
    const files = ['test1_worker.py', 'test2_worker.py']
    fs.readdir.mockReturnValueOnce(files)
    await expect(findWorkers('python_workers')).resolves.toEqual(files)

    // 2. Check empty directory
    fs.readdir.mockReturnValueOnce([])
    await expect(findWorkers('/')).resolves.toEqual([])

    // 3. Check argument error
    await expect(findWorkers()).rejects.toThrow(
      'Directory argument must be a string.'
    )

    // 4. Check non-existing path
    fs.readdir.mockImplementation(() => {
      throw new Error(
        `ENOENT: no such file or directory, scandir '/DOESNTEXIST'`
      )
    })
    await expect(findWorkers('/DOESNTEXIST')).rejects.toThrow(
      `ENOENT: no such file or directory, scandir '/DOESNTEXIST'`
    )
  })

  it('finds options', async () => {
    const getOptions = interfaces.__get__('getOptions')
    const options = [{ option1: 'value1' }]
    const workerProcess = new PythonWorker()

    // 1. Check the use case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: options } }))
    await expect(getOptions(workerProcess)).resolves.toEqual(options)
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 2. Check empty option case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: [] } }))
    await expect(getOptions(workerProcess)).resolves.toEqual([])
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 3. Check error case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { error: 'Timeout' } }))
    await expect(getOptions(workerProcess)).rejects.toThrow('Timeout')

    // 4. Throw error if workerProcess doesn't have a "send" property.
    expect(getOptions()).rejects.toThrow(
      'workerProcess not configured properly.'
    )
  })

  it('creates an interface', async () => {
    const createInterface = interfaces.__get__('createInterface')

    // 1. Check the use case
    mockSpawn.mockReturnValueOnce(AsyncMock({ result: { data: 'ready!' } }))
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: [] } }))
    fs.stat.mockReturnValueOnce({ mtime: now, size: 1234 })
    await expect(
      createInterface('python_workers', 'test1_worker.py', [])
    ).resolves.toMatchObject({
      interfaceName: 'test1',
      options: [],
      workerProcess: expect.objectContaining({
        send: expect.anything(),
        spawn: expect.anything()
      }),
      workerScript: expect.objectContaining({
        path: 'python_workers/test1_worker.py',
        mtime: now,
        updated: null
      })
    })
    expect(mockSpawn).toHaveBeenCalled()
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 2. Check that the update property is set for existing interfaces

    // 3. Check that existing interfaces are skipped

    // 4. Check argument error case
    await expect(createInterface()).rejects.toThrow(
      'Directory argument must be a string.'
    )

    // 5. Check missing file case
    fs.stat.mockImplementation(() => {
      throw new Error(
        `ENOENT: no such file or directory, scandir 'python_workers/testXXX_worker.py'`
      )
    })
    await expect(
      createInterface('python_workers', 'testXXX_worker.py', [])
    ).rejects.toThrow(
      `ENOENT: no such file or directory, scandir 'python_workers/testXXX_worker.py'`
    )
  })

  it('finds interfaces', async () => {
    const findInterfaces = interfaces.__get__('findInterfaces')

    const mockWorkerList = ['test1_worker.py', 'test2_worker.py']
    const mockInterfaceList = [
      { interfaceName: 'test1' },
      { interfaceName: 'test2' }
    ]
    const moduleWorkerdir = interfaces.__get__('WORKER_DIR')
    const mockFindWorkers = jest.fn()
    const mockCreateInterface = jest.fn()
    const revert = interfaces.__set__({
      findWorkers: mockFindWorkers,
      createInterface: mockCreateInterface
    })

    // 1. Check use case
    mockFindWorkers.mockReturnValueOnce(mockWorkerList)
    mockCreateInterface.mockReturnValueOnce(mockInterfaceList[0])
    mockCreateInterface.mockReturnValueOnce(mockInterfaceList[1])
    await expect(findInterfaces()).resolves.toEqual(mockInterfaceList)
    expect(mockFindWorkers).toHaveBeenCalledWith(moduleWorkerdir)
    expect(mockCreateInterface).toHaveBeenCalledWith(
      moduleWorkerdir,
      'test1_worker.py',
      expect.anything()
    )
    revert()
  })
})
