/**
 * @fileOverview INTERFACE BACKEND
 * @author <a href="tcve@u-blox.com">Tomi Cvetic</a>
 *
 * Communication with workers (for now only Python)
 * The workers are implemented as REPL clients that
 * communicate via JSON messages through STDIO.
 *
 * Convention:
 *  - Interface: Communication interface to instruments or DUTs (e.g. serial)
 *  - Worker: Python program that implements a REPL client (e.g. serial_worker.py)
 *  - Port: Port of an interface (e.g. COM1)
 *  - Connection: Process running the Python worker.
 */

// Module imports
const fs = require('fs')
const os = require('os')
const md5 = require('md5')
const { promisify } = require('util')
let PythonWorker = require('./pythonWorker')

const readdir = promisify(fs.readdir)
const stat = promisify(fs.stat)

/** @type {string} Directory with worker files */
const WORKER_DIR = `${__dirname}/python_workers`

/** @type {string} Hostname is used to identify ports across machines */
const HOST = os.hostname()

/**
 * Local state for the interface module
 * @typedef {object} State
 * @property {Interface[]} interfaces - Array of found interfaces
 * @property {Port[]} ports - Array of found ports
 * @property {Connection[]} connections - Array of established connections
 * @property {object} lastScan - Time of the last scan (per property)
 */
/** @type {State} Local state of the modules. */
const state = {
  interfaces: [], // @type {array}
  ports: [],
  connections: [],
  lastScan: {
    workers: Date.now() - 100000,
    ports: Date.now() - 100000,
    connections: Date.now() - 100000
  }
}

async function setState (content) {}

/** @type {string} GraphQL types for the interface module */
const typeDefs = `
  type Interface {
    id: ID!
    interfaceName: String!
    workerScript: WorkerScript!
    workerProcess: WorkerProcess
    ports: [Port]!
    connections: [Connection]!
    options: [Option!]
  }

  type WorkerScript {
    path: String!
    mtime: DateTime!
    updated: DateTime
  }

  type WorkerProcess {
    pid: Int!
    killed: Boolean!
    signalCode: String
    exitCode: Int
    spawnfile: String!
    spawnargs: [String]!
  }

  type Option {
    name: String!
    type: String!
    description: String
    values: [String!]
  }

  type Port {
    id: String!
    interfaceName: String!
    host: String!
    device: String
    name: String
    description: String
  }

  type Connection {
    id: ID!
    port: Port!
    workerProcess: WorkerProcess!
  }

  input ConnectionCommand {
    type: String!
    data: String
    options: String
  }

  extend type Query {
    interfaces: [Interface]!
    interface(id: ID!): Interface!
    ports: [Port]!
    port(id: ID!): Port!
    connections: [Connection]!
    connection(id: ID!): Connection!
  }

  extend type Mutation {
    connect(portId: ID!): Connection!
    endConnection(connectionId: ID!): Connection!
    killConnection(connectionId: ID!): Connection!
    sendCommand(connectionId: ID!, command: ConnectionCommand!): String!
  }
`

/**
 * Find worker files in a directory based on filename convention
 * (ending in _worker.py).
 * @param {string} directory - Python worker directory
 * @return {string[]} Array with found worker scripts
 */
async function findWorkers (directory) {
  // Only accept strings
  if (!(typeof directory === 'string')) {
    throw new Error('Directory argument must be a string.')
  }
  const fileNames = await readdir(directory)
  return fileNames.filter(fileName => fileName.includes('_worker.py'))
}

/**
 * Fetch options from Python worker process.
 * @typedef {object} workerProcess
 * @param {workerProcess} workerProcess - Worker process to use for fetching the options
 * @return {array} Option array
 */
async function getOptions (workerProcess) {
  // Check that workerProcess has a send method.
  if (!workerProcess || !workerProcess.send) {
    throw new Error('workerProcess not configured properly.')
  }
  const { data, error } = await workerProcess.send({ type: 'options' })
  if (error) throw new Error(error)
  return data
}

/**
 * Takes a Python worker script and generates an interface for it.
 * You need to call the spawn() function to launch the Python worker
 * process.
 * @typedef {object} WorkerScript
 * @property {string} path - Path of the worker script
 * @property {number} mtime - Last modification time
 * @property {number} updated - Last modification time, if changed after spawn
 *
 * @typedef {object} Interface
 * @property {string} id - ID of the interface
 * @property {string} interfaceName - Name of the interface
 * @property {WorkerScript} workerScript - Information about the worker script
 * @property {boolean} active - Is the worker still available
 *
 * @param {string} directory - Python worker directory
 * @param {string} workerFile - Python worker file
 * @param {Interface[]} interfaces - State with already connected interfaces
 * @return {Object} created interface
 */
async function createInterface (directory, workerFile, interfaces) {
  // Assert that arguments are strings.
  if (!(typeof directory === 'string' || directory instanceof String)) {
    throw new Error('Directory argument must be a string.')
  }
  if (!(typeof workerFile === 'string' || workerFile instanceof String)) {
    throw new Error('workerFile argument must be a string.')
  }
  const interfaceName = workerFile.replace(/_worker\.py/, '')
  const path = `${directory}/${workerFile}`
  // 1. Find out the last modification time.
  const { mtime } = await stat(path)
  const workerScript = { path, mtime, updated: null }
  // 2. Check if the interface already exists
  const iface = interfaces.find(iface => iface.interfaceName === interfaceName)
  if (iface) {
    // b. If it was modified, save the modification time.
    if (iface.workerScript.mtime < mtime) {
      iface.workerScript.update = mtime
    }
    return iface
  }
  // c. Spawn a new worker connection.
  const workerProcess = new PythonWorker(workerScript)
  const { data, error } = await workerProcess.spawn()
  if (error) throw new Error(error)
  const id = md5(
    `${HOST}${interfaceName}${workerScript.path}${mtime.toISOString()}`
  )
  const options = await getOptions(workerProcess)
  // d. Return the interface.
  return {
    id,
    interfaceName,
    workerScript,
    workerProcess,
    options,
    active: true
  }
}

/**
 * Generate interface list
 */
async function generateInterfaceList () {
  // 1. Don't check more frequently than once per second.
  // if (state.lastScan.workers + 1000 > Date.now()) return state.interfaces
  // state.lastScan.workers = Date.now()

  // 2. Find all files in ./python_workers that end in _worker.py
  const workerFiles = await findWorkers(WORKER_DIR)

  // 3. For every worker script
  const workerPromises = workerFiles.map(workerFile =>
    createInterface(WORKER_DIR, workerFile, state.interfaces)
  )
  return Promise.all(workerPromises)
}

async function interfaces (parent, args, context, info) {
  const incomingInterfaces = await generateInterfaceList()
  const newInterfaces = incomingInterfaces.filter(
    inIface => !state.interfaces.find(stIface => stIface.id === inIface.id)
  )
  state.interfaces.forEach(stIface => {
    if (!incomingInterfaces.find(inIface => inIface.id === stIface.id)) {
      stIface.active = false
    }
  })
  state.interfaces = [...state.interfaces, ...newInterfaces]
  return state.interfaces
}

async function iface (parent, { id }, context, info) {
  await generateInterfaceList()
  const iface = state.interfaces.find(
    iface => iface.id === id || iface.interfaceName === interfaceName
  )
  if (!iface) {
    throw new Error(`Worker id=${id} not found`)
  }
  return iface
}

/**
 * @typedef {object} workerProcess
 * @property {string} pid - OS process ID
 *
 * Serialize a worker process for GraphQL output
 * @param {*} parent
 * @param {*} args
 * @param {*} context
 * @param {*} info
 * @return {workerProcess} Serialized worker process
 */
function serializeWorkerProcess (parent, args, context, info) {
  const {
    killed,
    exitCode,
    signalCode,
    spawnargs,
    spawnfile,
    pid
  } = parent.workerProcess.pythonShell
  return {
    pid,
    killed,
    exitCode,
    signalCode,
    spawnfile,
    spawnargs
  }
}

async function createInterfacePorts (iface, ports) {
  // a) Ask interface for ports.
  const { data, error } = await iface.workerProcess.send({
    type: 'ports'
  })
  if (error) throw new Error(error)
  // b) Add all ports that are not in the list.
  return data.map(port => {
    const id = port.name || port.device
    return {
      id,
      interface: iface,
      interfaceName: iface.interfaceName,
      host: HOST,
      ...port
    }
  })
}

/**
 * Search for available ports on all interfaces
 * @typedef {object} Port
 * @property {string} id - Port ID
 * @property {string} interfaceName - Interface name
 * @property {host} host - Hostname
 * @property {string} device - Physical device name
 * @property {string} name - System name for the port
 * @property {string} description - Description if available by system
 *
 * @param {Port[]} ports - Array of already found ports
 * @return {Promise<Port[]>} Promise to return found ports
 */
async function generatePortList (ports) {
  // 1. Make sure, workers are updated.
  const incomingPorts = await generateInterfaceList()

  // 2. Don't check more frequently than once per second.
  // if (state.lastScan.ports + 1000 > Date.now()) return null
  // state.lastScan.ports = Date.now()

  // 3. Loop through all workers to find available ports.
  const portsPromises = interfaces.map(async iface => {
    createInterfacePorts(iface, state.ports)
  })
  await Promise.all(portsPromises)
}

async function ports (parent, args, context, info) {
  await generatePortList()

  if (parent) {
    return state.ports.filter(
      port => port.interfaceName === parent.interfaceName
    )
  } else {
    return state.ports
  }
}

async function port (parent, { id }, context, info) {
  await generatePortList()
  const port = state.ports.find(port => port.id === id)
  return port
}

/**
 * CONNECTION SECTION
 */
async function connect (parent, { portId }, ctx, info) {
  await generatePortList()
  const port = state.ports.find(port => port.id === portId)
  if (!port) throw new Error(`Port ${portId} not found`)

  const id = md5(port.interfaceName + port.device)
  if (state.connections.find(connection => connection.id === id)) {
    throw new Error('already connected.')
  }
  const workerProcess = new PythonWorker(port.interface.workerScript)
  const spawnData = await workerProcess.spawn()
  if (spawnData.error) throw new Error(spawnData.error)
  const connection = {
    id,
    port,
    workerProcess
  }
  const connectionData = await workerProcess.send({
    type: 'connect',
    device: port.device
  })
  if (connectionData.error) throw new Error(connectionData.error)
  state.connections.push(connection)
  return connection
}

async function connections (parent, args, ctx, info) {
  if (parent) {
    return state.connections.filter(
      connection => connection.port.interfaceName === parent.interfaceName
    )
  } else {
    return state.connections
  }
}

async function connection (parent, { id }, context, info) {
  const connection = state.connections.find(connection => connection.id === id)
  return connection
}

async function sendCommand (parent, { connectionId, command }, ctx, info) {
  const connection = state.connections.find(
    connection => connection.id === connectionId
  )
  if (!connection) throw new Error('Connection not found.')
  const { data, error } = await connection.workerProcess.send({ ...command })
  if (error) throw new Error(JSON.stringify(error))
  if (!data.response) {
    return JSON.stringify(data)
  } else {
    return data.response
  }
}

async function endConnection (parent, { connectionId }, ctx, info) {
  const connectionIndex = state.connections.findIndex(
    connection => connection.id === connectionId
  )
  if (connectionIndex < 0) throw new Error('Connection not found.')
  const connection = state.connections[connectionIndex]

  const { data, error } = await connection.workerProcess.end()
  if (error) throw new Error(JSON.stringify(error))
  if (data) throw new Error(`Process ended with exit code ${data}`)
  state.connections.splice(connectionIndex, 1)
  return connection
}

async function killConnection (parent, { connectionId, signal = 9 }, ctx, info) {
  const connectionIndex = state.connections.findIndex(
    connection => connection.id === connectionId
  )
  if (connectionIndex < 0) throw new Error('Connection not found.')
  const connection = state.connections[connectionIndex]

  const { data, error } = await connection.workerProcess.kill(signal)
  console.log({ data, error })
  if (error) throw new Error(JSON.stringify(error))
  if (data) throw new Error(`Received code ${data}`)
  state.connections.splice(connectionIndex, 1)
  return connection
}

const resolvers = {
  Query: {
    interfaces,
    interface: iface,
    ports,
    port,
    connections,
    connection
  },
  Mutation: {
    connect,
    sendCommand,
    endConnection,
    killConnection
  },
  Interface: {
    ports,
    connections,
    workerProcess: serializeWorkerProcess
  },
  Connection: {
    workerProcess: serializeWorkerProcess
  }
}

module.exports = { typeDefs, resolvers }
