from abc import ABCMeta, abstractmethod

class STATE:
    """
    Class containing all supported JTAG states.
    """
    RESET = 0
    IDLE = 1

class JTAG(metaclass=ABCMeta):
    """
    Abstract Class that defines a general JTAG interface.

    It offers four methods for usage:
    idle()
        Brings the JTAG state machine in the idle state
    reset()
        Brings the JTAG state machine to the reset state
    setIR(data, minLength=None)
        Sets the instruction register to *data*, where *data* is a string containing
        the binary data for the IR. The first character of *data* is shifted in first.

        If data is an unsigned integer, it is first converted to a binary string.

        *minLength* can be set to shift out data of a certain length. The IR data can be
        accessed in the objects *dataReg* attribute.

        The state machine will be in the idle state after this command.
    setDR(data, minLength=None)
        Sets the data register to *data*, where *data* is a string containing the
        binary data for the DR. The first character is shifted in last! So you can use MSB first
        notation.

        If data is an unsigned integer, it is first converted to a binary string.

        *minLength* can be set to shift out data of a certain length. The IR data can be
        accessed in the objects *dataReg* attribute.

        The state machine will be in the idle state after this command.

    Two methods need to be implemented by a child class, which implement the low level
    functionality of the JTAG interface:
    _shiftTAP(self, data)
        Needs to shift out data to the TAP.
        The function can assume, that a valid binary string is sent. It has to process the '1' and '0'
        character and ignore all other characters.
    _shiftReg(self, data)
        Needs to shift out data to a Register and record the response.
        The function can assume, that a valid binary string is sent. It has to process the '1' and '0'
        character and ignore all other characters.
    """

    state = STATE.RESET
    dataReg = ""
    instReg = ""

    def __init__(self, IRbefore=0, IRafter=0, DRbefore=0, DRafter=0, skipReset=False):
        """Constructor for the JTAG class."""

        # To be JTAG compliant, allow any number of instruction and data registers before and after the actual data.
        self.IRbefore = IRbefore
        self.IRafter = IRafter
        self.DRbefore = DRbefore
        self.DRafter = DRafter

        # Start with a clean reset state, unless requested differently
        if (not skipReset) is True:
            self.reset()

    def idle(self):
        """Brings the state machine to the idle state."""
        if STATE.RESET == self.state:
            self._shiftTAP("0")
        elif STATE.IDLE == self.state:
            pass

    def reset(self):
        """Brings the state machine to the reset state."""
        self._shiftTAP("11111111111111111111")
        self.state = STATE.RESET

    def setIR(self, data, minLen=None):
        """Sets the instruction register of the JTAG."""
        data = self.__checkData(data, minLen)
        if data is None:
            return
        if len(data) < minLen:
            data = "0" * (minLen - len(data)) + data

        # get safely into idle state before doing anything else.
        self.idle()
        # get into Shift-IR state
        self._shiftTAP("01100")
        # toggle in the before-IR data,
        out=self._shiftReg("1"*self.IRafter+data+"1"*self.IRbefore)
        # ... then the actual IR data
        #out = self._shiftReg(data)
        # ... and the after IR data.
        #self._shiftReg("1"*self.IRafter)
        # get into Pause-IR state
        self._shiftTAP("10")
        # store the shifted out IR data.
        out = out[::-1]
        out = out[self.IRafter:len(out)-self.IRbefore:]

        self.instReg = out
        return out

    def setDR(self, data, minLen=None):
        """Sets the data register of the JTAG."""
        data = self.__checkData(data, minLen)
        if data is None:
            return
        if len(data) < minLen:
            # pad data with zeros to fill minimum length.
            data += "0" * (minLen - len(data))

        # get safely into idle state before doing anything else.
        self.idle()
        # get into Shift-DR state
        self._shiftTAP("0100")
        # toggle in the before-DR data,
        #out=self._shiftReg("1"*self.DRafter+data+"1"*self.DRbefore)
        out=self._shiftReg("1"*self.DRafter+data+"1"*self.DRbefore)
        # ... then the actual DR data
        #out = self._shiftReg(data)
        # ... and the after DR data.
        #self._shiftReg("1"*self.IRafter)
        # get into Pause-DR state
        self._shiftTAP("10")
        # store the shifted out DR data.
        out = out[::-1]
        out = out[self.DRafter:len(out)-self.DRbefore]

        self.dataReg = out
        return out

    @staticmethod
    def __checkData(data, minLen):
        """Checks the data to be sent."""
        if isinstance(data, object):
            data = str(data)
        if isinstance(data, str):
            validSyms = "01 _'\t"
            lenValSyms = 0
            for sym in validSyms:
                lenValSyms += data.count(sym)
            if len(data) != lenValSyms:
                print("Error: Data is not a valid binary string. Data ignored.")
                return None
        elif isinstance(data, int) and data >= 0:
            data = bin(data)[2:]
        else:
            print("Error: Data needs to be a binary string or a non negative integer. Data ignored.")
            return None
        return data

    @abstractmethod
    def _shiftTAP(self, data):
        """Low level function to be implemented by the JTAG device driver."""
        pass

    @abstractmethod
    def _shiftReg(self, data):
        """Low level function to be implemented by the JTAG device driver."""
        pass
