import sys
import json
import serial
import serial.tools.list_ports
from utils import handle_exception


class SerialWorker:
    """
    Options offered by the PySerial module
    * https://pythonhosted.org/pyserial/pyserial_api.html
    * https://github.com/pyserial/pyserial/tree/master/serial
    """
    OPTIONS = [
        {"name": "port", "type": "String"},
        {"name": "baudrate", "type": "Int", "values": [
            50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200]},
        {"name": "bytesize", "type": "Int", "values": [5, 6, 7, 8]},
        {"name": "parity", "type": "String", "values": [
            "None", "Even", "Odd", "Mark", "Space"]},
        {"name": "timeout", "type": "Float"},
        {"name": "write_timeout", "type": "Float"},
        {"name": "xonxoff", "type": "Bool"},
        {"name": "rtscts", "type": "Bool"},
        {"name": "dsrdtr", "type": "Bool"},
        {"name": "inter_byte_timeout", "type": "Float"},
        {"name": "newline", "type": "String", "values": ["\r\n", "\n", ""]}
    ]

    connection = None
    newLine = "\r\n"

    @staticmethod
    def ports():
        driverPorts = serial.tools.list_ports.comports()

        ports = []

        for port in driverPorts:
            # https://pythonhosted.org/pyserial/tools.html#module-serial.tools.list_ports
            obj = {
                "device": port.device,
                "name": port.name,
                "description": port.description,
                "specific": {
                    "hwid": port.hwid,
                    "usb": {
                        "vid": port.vid,
                        "pid": port.vid,
                        "serial_number": port.serial_number,
                        "location": port.location,
                        "manufacturer": port.manufacturer,
                        "product": port.product,
                        "interface": port.interface
                    }
                }
            }
            ports.append(obj)

        return ports

    def connect(self, device, **kwargs):
        self.connection = serial.Serial(device, **kwargs)
        return {
            "device": device
        }

    def write(self, string):
        encodedString = (string + self.newLine).encode('ascii')
        self.connection.write(encodedString)
        return {
            "string": string,
            "encodedString": str(encodedString)
        }

    def read(self):
        original_response = self.connection.readline()
        response = original_response.decode('ascii').strip()
        garbage = []
        while self.connection.in_waiting:
            garbage.append(str(self.connection.readline()))

        return {
            "response": response,
            "originalResponse": str(original_response),
            "garbage": garbage
        }

    def ask(self, string):
        self.write(string)
        return self.read()

    def close(self):
        self.connection.close()
        return {}

    def __exit__(self, exc_type, exc_value, traceback):
        if self.connection is not None:
            self.connection.close()
            self.connection = None


worker = SerialWorker()
print(json.dumps({"data": "Serial ready"}), flush=True)

for line in sys.stdin:
    try:
        command = json.loads(line)
        if command['type'] == 'ports':
            res = SerialWorker.ports()
        elif command['type'] == 'options':
            res = SerialWorker.OPTIONS
        elif command['type'] == 'connect':
            options = command['options'] if 'options' in command else {}
            res = worker.connect(command['device'], **options)
        elif command['type'] == 'read':
            res = worker.read()
        elif command['type'] == 'write':
            res = worker.write(command['data'])
        elif command['type'] == 'ask':
            res = worker.ask(command['data'])
        print(json.dumps({"data": res}), flush=True)
    except Exception as error:
        print(handle_exception(error), flush=True)
