const { forwardTo } = require('prisma-binding')
const bcrypt = require('bcryptjs')
const jwt = require('jsonwebtoken')
const fs = require('fs')
const shortid = require('shortid')

const UPLOAD_DIR = './static/uploads'

const Query = {
  instruments: forwardTo('db'),
  instrument: (parent, args, context, info) => {
    const { id } = args
    return context.db.query.instrument({ where: { id } }, info)
  },
  projects: forwardTo('db'),
  projectVersions: forwardTo('db'),
  me: (parent, args, context, info) => {
    if (!context.request.userId) throw new Error('Not logged in.')
    return context.db.query.user(
      { where: { id: context.request.userId } },
      info
    )
  }
}

const storeFS = ({ stream, filename }) => {
  const id = shortid.generate()
  const path = `${UPLOAD_DIR}/${id}-${filename}`
  return new Promise((resolve, reject) =>
    stream
      .on('error', error => {
        if (stream.truncated) fs.unlink(path)
        reject(error)
      })
      .pipe(fs.createWriteStream(path))
      .on('error', error => reject(error))
      .on('finish', () => resolve({ id, path }))
  )
}

const processUpload = async upload => {
  const { createReadStream, filename, mimetype, encoding } = await upload
  const stream = createReadStream()
  let size = 0
  stream.on('data', chunk => {
    size += chunk.length
  })
  const { id, path } = await storeFS({ stream, filename })
  return { filename, mimetype, encoding, path, size }
}

const Mutation = {
  createUser: async (parent, args, context, info) => {
    const email = args.email.toLowerCase()
    const password = await bcrypt.hash(args.password, 10)
    console.log(email, password)
    const user = await context.db.mutation.createUser(
      {
        data: {
          ...args,
          email,
          password
        }
      },
      info
    )
    const token = jwt.sign({ userId: user.id }, process.env.APP_SECRET)
    context.response.cookie('token', token, {
      httpOnly: true,
      maxAge: 7 * 24 * 3600 * 1000
    })
    return user
  },
  userLogin: async (parent, args, context, info) => {
    const { email, password } = args
    const user = await context.db.query.user({ where: { email } })
    if (!user) throw new Error('User not found')
    const valid = await bcrypt.compare(password, user.password)
    if (!valid) throw new Error('Invalid password')
    const token = jwt.sign({ userId: user.id }, process.env.APP_SECRET)
    context.response.cookie(
      'token',
      token,
      {
        httpOnly: true,
        maxAge: 7 * 24 * 3600 * 1000
      },
      info
    )
    return user
  },
  userLogout: async (parent, args, context, info) => {
    context.response.clearCookie('token')
    return 'Logged out.'
  },
  createCharacterization: async (parent, args, context, info) => {
    const characterization = await context.db.mutation.createCharacterization(
      {
        data: {
          name: args.name,
          projectVersion: {
            connect: { id: args.projectVersion }
          }
        }
      },
      info
    )
    return characterization
  },
  createProject: async (parent, args, context, info) => {
    const project = await context.db.mutation.createProject(
      {
        data: { ...args }
      },
      info
    )
    return project
  },
  createProjectVersion: async (parent, args, context, info) => {
    console.log(args)
    const projectVersion = await context.db.mutation.createProjectVersion(
      {
        data: {
          ...args,
          date: new Date(args.date).toISOString(),
          project: {
            connect: {
              id: args.project
            }
          },
          changes: {
            set: args.changes
          }
        }
      },
      info
    )
    return projectVersion
  },
  uploadFiles: async (parent, args, context, info) => {
    console.log(args)
    const { name, description, file } = args
    let res
    for (const file of files) {
      res = await file
      console.log(res)
    }
    return res.filename
  },
  createInstrument: async (parent, args, context, info) => {
    const { interfaces, documents, commands } = args
    const { id, ...instrumentData } = args
    const files = []
    for (const document of documents) {
      const { file, id, ...formParams } = document
      const fileParams = await processUpload(file)
      files.push({ ...formParams, ...fileParams })
    }
    const dbCommands = commands.map(command => {
      const { id, ...dbCommand } = command
      return dbCommand
    })
    const instrument = await context.db.mutation.createInstrument(
      {
        data: {
          ...instrumentData,
          interfaces: { set: interfaces },
          documents: { create: files },
          commands: { create: dbCommands }
        }
      },
      info
    )
    return instrument
  }
}

const resolvers = {
  Query,
  Mutation
}

module.exports = { resolvers }
